# -*- coding: utf-8 -*-
"""3D volume discretization structure classes"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX, COLORS_DEFAULT
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools.geometry import get_step_from_geometry
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z


# --- base class volume discretized structure container - 3D
class StructDiscretized3D(StructBase):
    """base class volume discretized structure

    Using a list of positions and materials (for permittivites),
    this class defines the basic volume discretization, the
    polarizabilities and self-terms
    """

    __name__ = "3D volume discretized structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        materials,
        step=None,
        mesh_normalization_factor: float = 1,
        on_distance_violation: str = "warn",
        radiative_correction: bool = False,
        device: torch.device = None,
        **kwargs,
    ):
        """3D discretized structure

        Args:
            positions (torch.Tensor): meshpoint positions (3D, but all y values must be zero)
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            mesh_normalization_factor (float, optional): mesh normalization. Needs to be adapted for non-cubic meshes. Defaults to 1 (cubic).
            on_distance_violation (str, optional): behavior on distance violation. can be "error", "warn", None (silent), or "ignore" (do nothing, keep invalid meshpoints). Defaults to "error".
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to False.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        super().__init__(device=device, **kwargs)
        self.mesh = "3D"
        self.n_dim = 3

        # test for collisions:
        geo = torch.as_tensor(positions, dtype=DTYPE_FLOAT, device=self.device)

        if on_distance_violation.lower() == "ignore":
            geo_clean = geo
        if step is not None:
            norm = torch.norm(geo.unsqueeze(0) - geo.unsqueeze(1), dim=-1)
            norm[norm.triu() == 0] += 100 * step
            geo_clean = geo[norm.min(dim=0).values >= step * 0.999]
        else:
            warnings.warn("step not provided, cannot check mesh consistency.")
            geo_clean = geo

        if on_distance_violation.lower() == "error" and (len(geo) > len(geo_clean)):
            raise ValueError(
                "{} meshpoints in structure are too close!".format(
                    len(geo) - len(geo_clean)
                )
            )
        elif on_distance_violation.lower() == "warn" and (len(geo) > len(geo_clean)):
            warnings.warn(
                "{} meshpoints in structure are too close! Removing concerned meshpoints and continue.".format(
                    len(geo) - len(geo_clean)
                )
            )
        self.positions = torch.as_tensor(
            geo_clean, dtype=DTYPE_FLOAT, device=self.device
        )

        self.r0 = self.get_center_of_mass()  # center of gravity

        if step is None:
            step_scalar = get_step_from_geometry(self.positions)
        else:
            step_scalar = step
        # step for every meshcell, for consistency with other struct classes
        self.step = step_scalar * torch.ones(
            len(self.positions), dtype=DTYPE_FLOAT, device=self.device
        )

        self.mesh_normalization_factor = torch.as_tensor(
            mesh_normalization_factor, dtype=DTYPE_FLOAT, device=self.device
        )
        if mesh_normalization_factor == 1:
            self.mesh = "cubic"
        else:
            self.mesh = "hexagonal"

        self.radiative_correction = radiative_correction

        # material of each meshpoint
        if hasattr(materials, "__iter__"):
            if len(materials) != len(self.positions):
                raise ValueError(
                    "Either a global material needs to be given or "
                    + "each meshpoint needs a defined material. "
                    + "But meshpoint list and materials list are of different lengths."
                )
            self.materials = materials
        else:
            self.materials = [materials for i in self.positions]

        self.zeros = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )

        # discretized, made from natural material: only electric response
        self.interaction_type = "E"

    def __repr__(self, verbose=False):
        """description about structure"""
        from torchgdm.tools.misc import ptp

        out_str = ""
        out_str += "------ discretized 3D nano-object -------"
        out_str += "\n" + " mesh type:              {}".format(self.mesh)
        out_str += "\n" + " nr. of meshpoints:      {}".format(len(self.positions))
        out_str += "\n" + " nominal stepsizes (nm): {}".format(
            [float(f) for f in torch.unique(self.step)]
        )
        out_str += "\n" + " materials:              {}".format(
            [m.__name__ for m in set(self.materials)]
        )
        bnds = ptp(self.positions, dim=0)
        out_str += "\n" + " size & position:"
        out_str += "\n" + "     X-extension    :    {:.1f} (nm)".format(bnds[0])
        out_str += "\n" + "     Y-extension    :    {:.1f} (nm)".format(bnds[1])
        out_str += "\n" + "     Z-extension    :    {:.1f} (nm)".format(bnds[2])
        out_str += "\n" + "     center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
            *self.r0
        )

        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

        self.zeros = self.zeros.to(device)
        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)

        self.mesh_normalization_factor = self.mesh_normalization_factor.to(device)

        for mat in self.materials:
            mat.set_device(device)

    # --- self-terms
    def get_selfterm_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of 'EE' self-term tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pE self term tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )

        norm_nonrad = (
            (-4 * torch.pi)
            * self.mesh_normalization_factor
            / (3 * self.step**3 * eps_env)
        )

        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            norm_rad = (
                self.mesh_normalization_factor
                * (1j * k0**3 * (2 / 3))
                * torch.ones(
                    len(self.positions), dtype=DTYPE_COMPLEX, device=self.device
                )
            )
            cnorm = norm_nonrad + norm_rad
        else:
            cnorm = norm_nonrad

        self_terms_pE = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        self_terms_pE[:, 0, 0] = cnorm
        self_terms_pE[:, 1, 1] = cnorm
        self_terms_pE[:, 2, 2] = cnorm

        return self_terms_pE

    def get_selfterm_mE(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_selfterm_pH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_selfterm_mH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    # --- polarizabilities
    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of EE polarizability tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pE polarizability tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        eps_env_tensor = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        eps_env_tensor[:, 0, 0] = eps_env
        eps_env_tensor[:, 1, 1] = eps_env
        eps_env_tensor[:, 2, 2] = eps_env

        eps_geo = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        for i, mat in enumerate(self.materials):
            eps_geo[i] = mat.get_epsilon(wavelength)

        vcell_norm = self.step**3 / self.mesh_normalization_factor

        ## --- isotropic polarizability
        alpha_pE = (
            (eps_geo - eps_env_tensor)
            * vcell_norm.unsqueeze(1).unsqueeze(1)
            / (4.0 * torch.pi)
        )

        # with radiative reaction term:
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            alpha_pE = alpha_pE / (1 - (1j * k0**3 * (2 / 3)) * alpha_pE)

        return alpha_pE

    def get_polarizability_mE(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_polarizability_pH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_polarizability_mH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    # --- plotting
    def plot(
        self,
        projection="auto",
        scale=1.0,
        color="auto",
        show_grid=True,
        legend=True,
        set_ax_aspect=True,
        alpha=1.0,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the structure (2D)

        Args:
            projection (str, optional): Cartesian projection. Default: "XY" or plane in which all dipoles lie. Defaults to "auto".
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            alpha (int, optional): optional transparency. Defaults to 1.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d._plot_structure_discretized(
            self,
            projection=projection,
            scale=scale,
            color=color,
            show_grid=show_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot_contour(
        self,
        projection="auto",
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour around the structure (2D)

        Args:
            projection (str, optional): which cartesian plane to project onto. Defaults to "auto".
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d.contour(
            self,
            projection=projection,
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot3d(self, **kwargs):
        """plot the structure (3D)"""
        from torchgdm.visu import visu3d

        return visu3d.geo3d._plot_structure_discretized(self, **kwargs)

    # --- geometry operations
    def rotate(self, alpha, center=torch.as_tensor([0.0, 0.0, 0.0]), axis="z"):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis, one of ['x', 'y', 'z']. Defaults to 'z'.

        Raises:
            ValueError: unknown rotation axis

        Returns:
            :class:`StructDiscretized3D`: copy of structure with rotated geometry
        """
        _struct_rotated = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "x":
            rot = rotation_x(alpha, device=self.device)
        elif axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        elif axis.lower() == "z":
            rot = rotation_z(alpha, device=self.device)
        else:
            raise ValueError("Unknown rotation axis ''.".format(axis))

        if len(_struct_rotated.positions) > 1:
            _struct_rotated.positions = torch.matmul(
                _struct_rotated.positions - (center + self.r0), rot
            ) + (center + self.r0)
        else:
            warnings.warn("Single meshpoint found, ignore rotation.")

        _struct_rotated.r0 = _struct_rotated.get_center_of_mass()
        return _struct_rotated

    # --- effective model extraction wrapper
    def convert_to_effective_polarizability_pair(
        self,
        wavelengths: torch.Tensor,
        environment=None,
        test_accuracy: bool = False,
        only_pE_mH: bool = True,
        residual_warning_threshold=0.25,
        batch_size=16,
        **kwargs,
    ):
        """convert the structure to an effective point polarizability structure model

        The model consists of a pair of electric and magnetic dipole.

        kwargs are passed to :func:`extract_effective_polarizability`

        Args:
            wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm
            environment (environment class): 3D environement class. Defaults to None (vacuum).
            test_accuracy (bool, optional): Whether to test accuracy in a scattering simulation. Defaults to False.
            only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.
            residual_warning_threshold (float, optional). Maximum residual L2 norm before triggering a warning. Defaults to 0.25.

        Returns:
            :class:`torchgdm.struct.StructEffPola3D`: Effective dipole-pair model
        """
        from torchgdm.struct import StructEffPola3D
        from torchgdm.struct.eff_model_tools import extract_eff_pola_via_exact_mp_3d
        from torchgdm.struct.eff_model_tools import _test_effective_model_accuracy

        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)

        alpha = extract_eff_pola_via_exact_mp_3d(
            struct=self,
            wavelengths=wavelengths,
            environment=environment,
            only_pE_mH=only_pE_mH,
            batch_size=batch_size,
            residual_warning_threshold=residual_warning_threshold,
            **kwargs,
        )

        struct_aeff = StructEffPola3D(
            positions=torch.stack([self.r0]),
            alpha_dicts=[alpha],
            device=self.device,
        )

        # perform a test in an actual scattering simulation
        if test_accuracy:
            _test_results = _test_effective_model_accuracy(struct_aeff, self)

        return struct_aeff

    def convert_to_gpm(
        self,
        wavelengths,
        r_gpm,
        skeletonize=False,
        r_probe=None,
        illumination_list=None,
        environment=None,
        probe_type="particle",
        n_probe=1500,
        probe_spacing=3.0,
        n_planewave_sources=5,
        dipole_source_type="particle",
        n_dipole_sources=60,
        dipole_sources_spacing=5.0,
        verbose=True,
        progress_bar=True,
        device=None,
        residual_warning_threshold=0.1,
        test_accuracy=False,
        **kwargs,
    ):
        """convert the structure to a global polarizability matrix (GPM)

        Extract the GPM in a given `environement` at `wavelengths`.
        This is done in 3 steps:

            1) Illuminate with various sources, calculated scattered fields at
            various probe positions
            2) The effective dipole moment for each GPM dipole is obtained
            via matching of their emission and the probe fields of (1).
            3) A second inverse problem of adjusting the GPM to create the dipoles
            moments found in (2) is solved via pseudoinverse.

        By default, use a mix of plane waves and local dipole sources (different incidence directions, polarizations, locations, orientations).

        Args:
            struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
            wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            skeletonize (bool, optional): Has effect only if `r_gpm` is of type int. If True, perform a skeletonization prior clustering. Defaults to True.
            r_probe (torch.Tensor): probe positions where the simulated and the GPM scattered fields are matched. Overrides all other probe-related configurations. If not given, use automatic probe positions, which may not be optimal. Defaults to None (automatic).
            illumination_list (list of illuminations): List of torchgdm illumination fields to use for extraction. If not given, use automatic illumination fields, which may not be optimal. Defaults to None (automatic).
            environment (environment class): 3D environement class. Defaults to None (vacuum).
            probe_type (str, optional): Where to probe, one of ["particle", "sphere"]. "particle": probe at fixed distance to particle surface. "sphere": probe on enclosing sphere surface. Defaults to "particle".
            n_probe (int, optional): maximum number of probe positions on enclosing sphere. Defaults to 1500.
            probe_spacing (float, optional): additional distance to particle or to enclsing sphere surface, in units of discretization step. Defaults to 3.0.
            n_planewave_sources (int, optional): number of plane wave angles to use as illumination. Defaults to 7.
            dipole_source_type (str, optional): Where to put sources, one of ["particle", "sphere"]. "particle": probe at fixed distance to particle surface. "sphere": probe on enclosing sphere surface. Defaults to "particle".
            n_dipole_sources (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
            dipole_sources_spacing (float, optional): if using dipole light sources, additional distance to enclosing sphere surface, in units of discretization step. Defaults to 5.0.
            verbose (bool, optional): whether to sprint progess info. Defaults to True.
            progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
            device (str, optional). Defaults to None, in which case the structure's device is used.
            residual_warning_threshold (float, optional). Maximum residual L2 norm before triggering a warning. Defaults to 0.1.
            test_accuracy (bool, optional): Whether to test accuracy in a scattering simulation. Defaults to False.

        Returns:
            :class:`torchgdm.struct.struct3d.gpm.StructGPM3D`: Effective dipole-pair model
        """
        from torchgdm.struct.struct3d.gpm3d import StructGPM3D
        from torchgdm.struct.eff_model_tools import extract_gpm_from_struct
        from torchgdm.struct.eff_model_tools import _test_effective_model_accuracy

        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)

        gpm_dict = extract_gpm_from_struct(
            struct=self,
            wavelengths=wavelengths,
            r_gpm=r_gpm,
            skeletonize=skeletonize,
            r_probe=r_probe,
            illumination_list=illumination_list,
            environment=environment,
            probe_type=probe_type,
            n_probe=n_probe,
            probe_spacing=probe_spacing,
            n_planewave_sources=n_planewave_sources,
            dipole_source_type=dipole_source_type,
            n_dipole_sources=n_dipole_sources,
            dipole_sources_spacing=dipole_sources_spacing,
            verbose=verbose,
            progress_bar=progress_bar,
            device=device,
            residual_warning_threshold=residual_warning_threshold,
            **kwargs,
        )

        struct_gpm = StructGPM3D(
            positions=self.r0,
            gpm_dicts=gpm_dict,
            device=self.device,
            progress_bar=progress_bar,
        )

        # perform a test in an actual scattering simulation
        if test_accuracy:
            _test_results = _test_effective_model_accuracy(
                struct_gpm, self, which=["ecs", "nf_sca"]
            )

        return struct_gpm


class StructDiscretizedCubic3D(StructDiscretized3D):
    """class for cubic volume discretized structure

    Defines the cubic volume discretization, polarizabilities and self-terms
    """

    __name__ = "3D cubic discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs,
    ):
        """3D structure, discretized on a cubic lattice

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        from torchgdm.struct.struct3d import discretizer_cubic

        positions = discretizer_cubic(*discretization_config, step=step)

        super().__init__(
            positions,
            materials,
            step=step,
            device=device,
            on_distance_violation="ignore",  # trust the discretizer --> speedup
            **kwargs,
        )

        self.mesh = "cubic"
        self.mesh_normalization_factor = torch.tensor(
            1.0, dtype=DTYPE_FLOAT, device=self.device
        )


class StructDiscretizedHexagonal3D(StructDiscretized3D):
    """class for hexagonal compact volume discretized structure

    Defines the hexagonal compact volume discretization, polarizabilities and self-terms
    """

    __name__ = "3D hexagonal compact discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs,
    ):
        """3D structure, discretized on a hexagonal compact lattice

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        from torchgdm.struct.struct3d import discretizer_hexagonalcompact

        positions = discretizer_hexagonalcompact(*discretization_config, step=step)

        super().__init__(
            positions,
            materials,
            step=step,
            device=device,
            on_distance_violation="ignore",  # trust the discretizer --> speedup
            **kwargs,
        )

        self.mesh = "hexagonal"
        self.mesh_normalization_factor = torch.sqrt(
            torch.as_tensor(2.0, dtype=DTYPE_FLOAT, device=self.device)
        )



# encoding: utf-8
"""
unittests for global polarizability matrix (GPM)

todo:
  - test accuracy (vs Mie?)
  - test combine of GPM structures

author: P. Wiecha, 04/2025
"""

# %%
import warnings
import unittest

import torch
import numpy as np

# deterministinc randomness (reproducible eff. model extraction)
torch.manual_seed(42)

import torchgdm as tg

SEED = 123


class TestSim3DSplitCombine(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("test splitting and combining simulations...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        self.wls = torch.linspace(450.0, 750.0, 2)
        self.env = tg.env.EnvHomogeneous2D(env_material=1.0)
        self.inc = tg.env.freespace_2d.PlaneWave(
            e0p=0.71, e0s=0.71, inc_angle=torch.pi / 6
        )

        # - discretized structure
        self.struct = tg.struct2d.StructDiscretizedSquare2D(
            discretization_config=tg.struct2d.rectangle(l=25, h=10),
            step=10,
            materials=tg.materials.MatConstant(eps=16),
        )

    def test_gpm_vs_effpola_model(self):
        for device in self.devices:
            # small structure that with dipolar response
            struct = tg.struct2d.StructDiscretizedSquare2D(
                discretization_config=tg.struct2d.rectangle(l=5, h=4),
                step=10,
                materials=tg.materials.MatConstant(eps=16),
            )
            struct.set_device(device)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_eff1 = struct.convert_to_effective_polarizability_pair(
                    self.wls,
                    environment=self.env,
                    verbose=False,
                    progress_bar=False,
                    test_accuracy=False,
                    residual_warning_threshold=1000,
                )

                n_gpm_dp = 1
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm1 = struct.convert_to_gpm(
                    self.wls,
                    n_gpm_dp,
                    environment=self.env,
                    verbose=False,
                    progress_bar=False,
                    test_accuracy=False,
                    residual_warning_threshold=1000,
                )

            sim_conf = dict(
                illumination_fields=self.inc, environment=self.env, wavelengths=self.wls
            )
            sim_gpm = tg.Simulation(structures=s_gpm1, **sim_conf)
            sim_gpm.run(verbose=False, progress_bar=False)

            sim_effdp = tg.Simulation(structures=s_eff1, **sim_conf)
            sim_effdp.run(verbose=False, progress_bar=False)

            wl = self.wls[0]
            r_probe = tg.tools.geometry.coordinate_map_2d_square(1000, 5, r3=100)

            nf_eff = sim_effdp.get_nearfield(wl, r_probe, progress_bar=False)["sca"]
            nf_gpm = sim_gpm.get_nearfield(wl, r_probe, progress_bar=False)["sca"]

            # single dipole GPM similar should give very result as eff-dipole extraction
            # print((nf_eff.efield).abs())
            # print((nf_gpm.efield).abs())
            # print((nf_eff.efield - nf_gpm.efield).abs().mean())
            assert (nf_eff.efield - nf_gpm.efield).abs().mean() < 0.001

    def test_gpm_translation(self):
        for device in self.devices:

            struct_a = self.struct.copy()
            struct_a.set_device(device)

            shift = [200, 0, 200]
            struct_b = self.struct.copy()
            struct_b.set_device(device)
            struct_b = struct_b + shift  # shift the discretized structure

            # manually define effective dipole positions
            r_gpm_0 = torch.as_tensor(
                [[float(x), 0, 0] for x in np.linspace(-100, 100, 5)], device=device
            )

            conf = dict(
                wavelengths=self.wls,
                verbose=False,
                environment=self.env,
                progress_bar=False,
                test_accuracy=False,
                residual_warning_threshold=1000,
                n_dipole_sources=0,  # optimize far-field illumination for testing
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                r_gpm_a = r_gpm_0 + struct_a.r0
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm_a = struct_a.convert_to_gpm(r_gpm=r_gpm_a, **conf)
                s_gpm_a = s_gpm_a + shift  # shift the GPM structure

                r_gpm_b = r_gpm_0 + struct_b.r0
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                # convert the shifted discretization
                s_gpm_b = struct_b.convert_to_gpm(r_gpm=r_gpm_b, **conf)

                # run simulations
                sim_conf = dict(
                    illumination_fields=self.inc,
                    environment=self.env,
                    wavelengths=self.wls,
                )
                sim_gpm_a = tg.Simulation(structures=s_gpm_a, **sim_conf)
                sim_gpm_b = tg.Simulation(structures=s_gpm_b, **sim_conf)
                sim_gpm_a.run(verbose=False, progress_bar=False)
                sim_gpm_b.run(verbose=False, progress_bar=False)

                # compare fields
                r_probe = tg.tools.geometry.coordinate_map_2d_square(500, 3, r3=800)
                for wl in self.wls:
                    nf1 = sim_gpm_a.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    nf2 = sim_gpm_b.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    # print(nf1.efield)
                    # print(nf2.efield)
                    # print((nf1.efield-nf2.efield).abs())
                    torch.testing.assert_close(
                        nf1.efield, nf2.efield, atol=1e-4, rtol=1e-5
                    )
                    torch.testing.assert_close(
                        nf1.hfield, nf2.hfield, atol=1e-4, rtol=1e-5
                    )

    def test_gpm_rotation(self):
        for device in self.devices:

            struct_a = self.struct.copy()
            struct_a.set_center_of_mass([0, 0, 0])
            struct_a.set_device(device)

            rotation_angle = torch.pi / 2
            struct_b = self.struct.copy()
            struct_b.set_center_of_mass([0, 0, 0])
            struct_b.set_device(device)
            struct_b = struct_b.rotate(
                rotation_angle, axis="y"
            )  # rotate the discretized struct

            # manually define effective dipole positions
            r_gpm = torch.as_tensor(
                [[float(x), 0, 0] for x in np.linspace(-100, 100, 10)], device=device
            )

            conf = dict(
                wavelengths=self.wls,
                verbose=False,
                environment=self.env,
                progress_bar=False,
                test_accuracy=False,
                residual_warning_threshold=1000,
                n_dipole_sources=0,  # optimize far-field illumination for testing
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                r_gpm_a = r_gpm + struct_a.r0
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm_a = struct_a.convert_to_gpm(r_gpm=r_gpm_a, **conf)
                s_gpm_a = s_gpm_a.rotate(
                    rotation_angle, axis="y"
                )  # rotate the GPM struct

                # conv rotated discr. struct
                r_gpm_b = r_gpm + struct_b.r0
                rot_y = tg.tools.geometry.rotation_y(rotation_angle).to(device)
                r_gpm_b = torch.matmul(r_gpm_b, rot_y)
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm_b = struct_b.convert_to_gpm(r_gpm=r_gpm_b, **conf)

                # run simulations
                sim_conf = dict(
                    illumination_fields=self.inc,
                    environment=self.env,
                    wavelengths=self.wls,
                )
                sim_gpm_a = tg.Simulation(structures=s_gpm_a, **sim_conf)
                sim_gpm_b = tg.Simulation(structures=s_gpm_b, **sim_conf)
                sim_gpm_a.run(verbose=False, progress_bar=False)
                sim_gpm_b.run(verbose=False, progress_bar=False)

                # compare fields
                r_probe = tg.tools.geometry.coordinate_map_2d_square(
                    600, 3, r3=800, projection="xy"
                )
                for wl in self.wls:
                    nf1 = sim_gpm_a.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    nf2 = sim_gpm_b.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    # print(nf1.efield.abs())
                    # print(nf2.efield.abs())
                    # print((nf1.efield-nf2.efield).abs())
                    torch.testing.assert_close(
                        nf1.efield, nf2.efield, atol=0.005, rtol=0.5
                    )
                    torch.testing.assert_close(
                        nf1.hfield, nf2.hfield, atol=0.005, rtol=0.5
                    )

    def test_gpm_combine(self):
        for device in self.devices:

            struct_a = self.struct.copy()
            struct_a.set_device(device)

            struct_b = self.struct.copy()
            struct_b.set_device(device)

            # manually define effective dipole positions
            r_gpm_0 = torch.as_tensor(
                [[float(x), 0, 0] for x in np.linspace(-100, 100, 10)], device=device
            )

            conf = dict(
                wavelengths=self.wls,
                verbose=False,
                environment=self.env,
                progress_bar=False,
                test_accuracy=False,
                residual_warning_threshold=1000,
                n_dipole_sources=0,  # optimize far-field illumination for testing
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                r_gpm_a = r_gpm_0 + struct_a.r0
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm_a = struct_a.convert_to_gpm(r_gpm=r_gpm_a, **conf)
                s_gpm_a = s_gpm_a + [200, 0, 200]

                r_gpm_b = r_gpm_0 + struct_b.r0
                torch.manual_seed(SEED)
                np.random.seed(SEED)
                s_gpm_b = struct_b.convert_to_gpm(r_gpm=r_gpm_b, **conf)

                # run simulations
                sim_conf = dict(
                    illumination_fields=self.inc,
                    environment=self.env,
                    wavelengths=self.wls,
                )
                sim_gpm_a = tg.Simulation(structures=[s_gpm_a, s_gpm_b], **sim_conf)
                sim_gpm_b = tg.Simulation(structures=s_gpm_a + s_gpm_b, **sim_conf)
                sim_gpm_a.run(verbose=False, progress_bar=False)
                sim_gpm_b.run(verbose=False, progress_bar=False)

                # compare fields
                r_probe = tg.tools.geometry.coordinate_map_2d_square(1000, 3, r3=500)
                for wl in self.wls:
                    nf1 = sim_gpm_a.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    nf2 = sim_gpm_b.get_nearfield(wl, r_probe, progress_bar=False)[
                        "sca"
                    ]
                    # print(nf1.efield)
                    # print(nf2.efield)
                    # print((nf1.efield-nf2.efield).abs())
                    torch.testing.assert_close(
                        nf1.efield, nf2.efield, atol=1e-4, rtol=1e-5
                    )
                    torch.testing.assert_close(
                        nf1.hfield, nf2.hfield, atol=1e-4, rtol=1e-5
                    )


# %%
if __name__ == "__main__":
    print("testing 2D global polarizability matrix manipulations...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""package for torchgdm structures

Contains Classes for 3D and 2D discretized structures and point (3D) and line (2D) effective polarizability structures:

.. currentmodule:: torchgdm.struct

3D
--

.. autosummary::
   :toctree: generated/
   
   StructDiscretizedCubic3D
   StructDiscretizedHexagonal3D
   StructGPM3D
   StructMieSphereGPM3D
   StructTMatrixGPM3D


dipolar approximation
^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   :toctree: generated/
   
   StructEffPola3D
   StructMieSphereEffPola3D


3D discretization
-----------------

.. autosummary::
   :toctree: generated/
   
   struct3d


2D
--

.. autosummary::
   :toctree: generated/
   
   StructDiscretizedSquare2D
   StructGPM2D
   StructMieCylinderGPM2D
   StructTMatrixGPM2D


dipolar approximation
^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   :toctree: generated/
   
   StructEffPola2D
   StructMieCylinderEffPola2D


2D discretization
-----------------

.. autosummary::
   :toctree: generated/
   
   struct2d

base class
----------

.. autosummary::
   :toctree: generated/
   
   StructBase

"""
# - packages / modules
from . import struct2d
from . import struct3d
from . import eff_model_tools


# - 3D
from .struct3d.volume import StructDiscretized3D
from .struct3d.volume import StructDiscretizedCubic3D
from .struct3d.volume import StructDiscretizedHexagonal3D

from .struct3d.point import StructMieSphereEffPola3D  # Mie core-shell sphere
from .struct3d.point import StructEffPola3D

from .struct3d.gpm3d import StructGPM3D
from .struct3d.gpm3d import StructMieSphereGPM3D
from .struct3d.gpm3d import StructTMatrixGPM3D

# - 2D
from .struct2d.surface import StructDiscretized2D
from .struct2d.surface import StructDiscretizedSquare2D

from .struct2d.line import StructMieCylinderEffPola2D  # Mie core-shell cylinder
from .struct2d.line import StructEffPola2D

from .struct2d.gpm2d import StructGPM2D
from .struct2d.gpm2d import StructMieCylinderGPM2D
from .struct2d.gpm2d import StructTMatrixGPM2D

# - base class
from .base_classes import StructBase



# -*- coding: utf-8 -*-
"""3D volume discretizations

.. currentmodule:: torchgdm.struct.struct3d

Classes
-------

.. autosummary::
   :toctree: generated/
   
   StructDiscretized3D
   StructDiscretizedCubic3D
   StructDiscretizedHexagonal3D
   StructGPM3D
   StructMieSphereGPM3D
   StructTMatrixGPM3D


dipolar approximation
^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   :toctree: generated/
   
   StructEffPola3D
   StructMieSphereEffPola3D


Functions
---------

.. autosummary::
   :toctree: generated/
   
   extract_gpm_from_struct
   extract_gpm_from_tmatrix
   extract_eff_pola_via_exact_mp_3d
   get_gpm_positions_by_clustering


Geometries
----------

.. autosummary::
   :toctree: generated/
   
   cube
   cuboid
   sphere
   spheroid
   disc
   ellipse
   split_ring
   lshape
   prism_trigonal
   from_image


Discretizer functions
---------------------

.. autosummary::
   :toctree: generated/
   
   discretizer_cubic
   discretizer_hexagonalcompact


"""
from . import volume
from . import point
from . import gpm3d
from . import geometries

from .volume import StructDiscretizedCubic3D
from .volume import StructDiscretizedHexagonal3D
from .volume import StructDiscretized3D
from .point import StructEffPola3D
from .point import StructMieSphereEffPola3D
from .gpm3d import StructGPM3D
from .gpm3d import StructMieSphereGPM3D
from .gpm3d import StructTMatrixGPM3D

from torchgdm.struct.eff_model_tools import extract_eff_pola_via_exact_mp_3d
from torchgdm.struct.eff_model_tools import extract_gpm_from_struct
from torchgdm.struct.eff_model_tools import extract_gpm_from_tmatrix
from torchgdm.struct.eff_model_tools import get_gpm_positions_by_clustering

from .geometries import discretizer_cubic
from .geometries import discretizer_hexagonalcompact

from .geometries import cube
from .geometries import cuboid
from .geometries import sphere
from .geometries import spheroid
from .geometries import disc
from .geometries import ellipse
from .geometries import split_ring
from .geometries import lshape
from .geometries import prism_trigonal

from .geometries import from_image



"""
global polarizability matrix (GPM) module

see:
Bertrand, M., Devilez, A., Hugonin, J.-P., Lalanne, P. & Vynck, K.
*Global polarizability matrix method for efficient modeling of light scattering by dense ensembles of non-spherical particles in stratified media.*
JOSA A 37, 70-83 (2020)
DOI: 10.1364/JOSAA.37.000070

author: P. Wiecha, 03/2025
"""

import warnings

import torch
from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools.geometry import get_enclosing_sphere_radius
from torchgdm.tools.geometry import test_structure_distances
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z
from torchgdm.tools import interp
from torchgdm.tools.misc import ptp


class StructGPM3D(StructBase):
    """3D global polarizability matrix structure

    Defines a GPM-based structure

    for details about GPM, see:

    Bertrand, M., Devilez, A., Hugonin, J.-P., Lalanne, P. & Vynck, K.
    *Global polarizability matrix method for efficient modeling of light scattering by dense ensembles of non-spherical particles in stratified media.*
    JOSA A 37, 70-83 (2020), DOI: 10.1364/JOSAA.37.000070
    """

    __name__ = "global polarizability matrix 3D structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        gpm_dicts: list,
        radiative_correction: bool = True,
        device: torch.device = None,
        environment=None,
        shift_z_to_r0: bool = True,
        progress_bar=True,
    ):
        """3D global polarizability matrix (GPM) class

        The main information is provided in the `gpm_dicts` argument, which is a list of dicts with the full global polarizability matrix definitions. Each dict defines one structure and must contain following:
            - 'wavelengths': wavelengths at which the polarizabilities are calculated
            - 'GPM_N6xN6':
                6N x 6N global polarizability tensors, each of shape [len(wavelengths),6N,6N]. Each 6x6 sub-matrix describes electric and magnetic dipole moments induced by and electric and magentic field.
                N is the number of dipole pairs for the GPM. Caution! All GPMs in one structure must have the same number of dipoles.
            optional keys:
            - 'full_geometry': the original volume discretization of the represented geometry
            - 'r0': the origin of the effective polarizabilities with respect to optional 'full_geometry'
            - 'enclosing_radius': enclosing radius of the original structure

        Args:
            positions (torch.Tensor): positions of the individual GPMs (same size as `gpm_dicts`)
            gpm_dicts (list): list of polarizability model dictionaries (same size as `positions`)
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to True.
            device (torch.device, optional): Defaults to "cpu".
            environment (environment instance, optional): 3D environment class. Defaults to None.
            shift_z_to_r0 (bool, optional): If True, if a position z-value is zero, each polarizability model's z position will be shifted to the height of the effective dipole development center. Defaults to True.
            progress_bar (bool optional): whether to show progress bars on internal compute. Defaults to True.

        Raises:
            ValueError: _description_
        """
        super().__init__(device=device)

        self.n_dim = 3
        self.interaction_type = "EH"
        self.gpm_dict = gpm_dicts

        # expand positions, put single scatterer in list
        move_positions = torch.as_tensor(
            positions, dtype=DTYPE_FLOAT, device=self.device
        )
        move_positions = torch.atleast_2d(move_positions)
        assert move_positions.shape[1] == 3

        self.radiative_correction = radiative_correction

        # single alpha_dict: put in list
        if type(gpm_dicts) == dict:
            gpm_dicts = [gpm_dicts] * len(move_positions)
        for _gd in gpm_dicts:
            assert type(_gd) == dict

        # environment at which alpha has been extracted (if given):
        if environment is None:
            self.environment = gpm_dicts[0]["environment"]
        else:
            warnings.warn("Using different environment than specified in GPM-dict.")
            self.environment = environment
        self.progress_bar = progress_bar

        for _adict in gpm_dicts:
            assert "wavelengths" in _adict
            if "GPM_N6xN6" not in _adict:
                raise ValueError(
                    "Global polarizability matrix description dicts must contain GPM tensors "
                    + "under the dict key: 'GPM_N6xN6'."
                )
            if "r_gpm" not in _adict:
                raise ValueError(
                    "Global polarizability matrix description dicts must contain GPM position tensors "
                    + "under the dict key: 'r_gpm'."
                )

        # use first pola-tensor wavelength range
        wavelengths = gpm_dicts[0]["wavelengths"]
        n_dp_gpm = len(gpm_dicts[0]["r_gpm"])
        self.wavelengths_data = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )

        for _adict in gpm_dicts:
            _wls = torch.as_tensor(
                _adict["wavelengths"], dtype=DTYPE_FLOAT, device=self.device
            )
            if len(_adict["r_gpm"]) != n_dp_gpm:
                raise ValueError(
                    "All GPMs must have the same number of dipoles, if several GPMs are combined in one structure."
                )
            if not torch.all(torch.isin(self.wavelengths_data, _wls)):
                warnings.warn(
                    "Pre-calculated wavelengths of the structures are not identical. "
                    + "Try to use linear interpolation to fill missing values. "
                    + "This may be inaccurate."
                )
            if (torch.min(_wls) > torch.min(self.wavelengths_data)) or (
                torch.max(_wls) < torch.max(self.wavelengths_data)
            ):
                raise ValueError(
                    "Interpolation not possible. Pre-calculated wavelengths of the structures "
                    + "must be within the same range to allow interpolation."
                )

        # optionally shift z such that 'bottom' is at z=0. Do only if z=0
        if shift_z_to_r0:
            for i, p in enumerate(move_positions):
                if p[2] == 0:
                    p[2] = torch.as_tensor(
                        gpm_dicts[i]["r0"][2], dtype=p.dtype, device=p.device
                    )

        # populate polarizability tensors database for each position at
        # pre-calculated wavelengths. if necessary, interpolate between data points
        gpm_pos = []
        gpm_data = []
        reduced_step = []
        full_geometries = []
        for i, _adict in enumerate(gpm_dicts):
            gpm_pos.append(
                _adict["r_gpm"] - _adict["r0"] + move_positions[i].unsqueeze(0)
            )
            if torch.all(_adict["wavelengths"] == self.wavelengths_data):
                _gpm = _adict["GPM_N6xN6"]
            else:
                # if requested other wavelengths than given, interpolate
                _gpm = self._interpolate_single_alpha(
                    wls=self.wavelengths_data,
                    gpm_data=_adict["GPM_N6xN6"],
                    wl_data=_adict["wavelengths"],
                )
            gpm_data.append(_gpm)

            # optional values
            if "full_geometry" in _adict:
                full_geometries.append(
                    torch.as_tensor(
                        _adict["full_geometry"] - _adict["r0"],
                        dtype=DTYPE_FLOAT,
                        device=self.device,
                    )
                    + move_positions[i]
                )

            if "enclosing_radius" in _adict:
                _rstep = _adict["enclosing_radius"] * 2
            else:
                if "full_geometry" in _adict:
                    _r_eff = get_enclosing_sphere_radius(full_geometries[-1])
                    _rstep = _r_eff
                else:
                    _rstep = 0
            reduced_step += [_rstep / n_dp_gpm**0.66] * n_dp_gpm

        self.full_geometries = full_geometries

        # effective polarizabilities of each meshpoint at each wavelength: shape (Npos, Nwl, 6, 6)
        if len(gpm_data) != 0:
            self.gpm_data = torch.stack(gpm_data)
            self.gpm_pos = torch.stack(gpm_pos)
        else:
            raise ValueError("Unexpected error: Not GPM data.")

        # selfterms are zero. This acts as a placeholder and is not being used
        self.selfterms_data = torch.zeros_like(self.gpm_data)

        # populate lookup tables
        self.create_lookup()

        # other parameters ("step" corresponds to the effective diameter divided by nr of GPM dipole pairs)
        self.step = torch.as_tensor(reduced_step, dtype=DTYPE_FLOAT, device=self.device)

        # center of gravity of ensemble of all points
        self.r0 = self.get_center_of_mass()

        # set device
        self.set_device(self.device)

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ 3D Global Polarizability Matrix nano-object -------"
        out_str += "\n" + " nr. of GPM structures:  {}".format(self.gpm_pos.shape[0])
        out_str += "\n" + " nr. of dipoles per GPM: {} (each {} ED and MD)".format(
            self.gpm_pos.shape[1] * 2, self.gpm_pos.shape[1]
        )
        out_str += "\n" + " total nr. of dipoles:   {} (each {} ED and MD)".format(
            self.gpm_pos.shape[0] * self.gpm_pos.shape[1] * 2,
            self.gpm_pos.shape[0] * self.gpm_pos.shape[1],
        )
        if len(self.full_geometries) > 0:
            pos = torch.cat(self.full_geometries)
            out_str += "\n" + " original geometry: "
            out_str += "\n" + "  - replacing nr. of meshpoints: {}".format(len(pos))
            bnds = ptp(pos, dim=0)
            out_str += "\n" + "  - size & position:"
            out_str += "\n" + "        X-extension    : {:.1f} (nm)".format(bnds[0])
            out_str += "\n" + "        Y-extension    : {:.1f} (nm)".format(bnds[1])
            out_str += "\n" + "        Z-extension    : {:.1f} (nm)".format(bnds[2])
            out_str += "\n" + "  - center of mass : ({:.1f}, {:.1f}, {:.1f})".format(
                *[float(f) for f in self.get_center_of_mass()]
            )

        return out_str

    def get_all_positions(self) -> torch.Tensor:
        return torch.reshape(self.gpm_pos, (-1, 3))

    def get_r_pm(self):
        """positions of electric and magnetic polarizable dipoles"""
        r_p = self.get_all_positions()
        r_m = self.get_all_positions()
        return r_p, r_m

    def get_e_selfconsistent(self, wavelength, **kwargs):
        if float(wavelength) not in self.fields_inside:
            raise ValueError(
                f"Inside field not available at wl={wavelength}nm. "
                + "Run the simulation."
            )
        return self.fields_inside[float(wavelength)].get_efield()

    def get_h_selfconsistent(self, wavelength, **kwargs):
        if float(wavelength) not in self.fields_inside:
            raise ValueError(
                f"Inside field not available at wl={wavelength}nm. "
                + "Run the simulation."
            )
        return self.fields_inside[float(wavelength)].get_hfield()

    def get_e_h_selfconsistent(self, wavelength, **kwargs) -> torch.Tensor:
        e = self.get_e_selfconsistent(wavelength)
        h = self.get_h_selfconsistent(wavelength)
        return torch.cat([e, h], dim=-1)

    def get_pm(self, wavelength):
        """self-consistent electric and magnetic dipole moments"""
        gpm = self.get_gpm(wavelength, self.environment)
        f_in = self.get_e_h_selfconsistent(wavelength)
        f_per_gpm = f_in.chunk(chunks=len(self.gpm_pos), dim=1)
        p = []
        m = []
        for i, _f in enumerate(f_per_gpm):
            _f = _f.reshape(len(_f), -1)
            _pm = torch.matmul(gpm[i], _f.unsqueeze(-1))[..., 0]
            _pm = _pm.reshape(len(_pm), -1, 6)
            p.append(_pm[..., :3])
            m.append(_pm[..., 3:])

        p = torch.cat(p, dim=1)
        m = torch.cat(m, dim=1)

        return p, m

    def get_center_of_mass(self) -> torch.Tensor:
        # use full geometries if available
        if len(self.full_geometries) == len(self.gpm_pos):
            r0 = [_pos.mean(dim=0) for _pos in self.full_geometries]
            r0 = torch.stack(r0, axis=0).mean(dim=0)
        else:
            r0 = self.gpm_pos.mean(dim=(0, 1))
        return r0

    def create_lookup(self):
        """Create a lookup table for the polarizability tensors"""
        # populate lookup tables with pre-calculated data
        self.lookup = {}
        for i_wl, wl in enumerate(self.wavelengths_data):
            self.lookup[round(float(wl), 3)] = self.gpm_data[:, i_wl, :, :]

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)
        if self.environment is not None:
            self.environment.set_device(device)

        if len(self.full_geometries) > 0:
            self.full_geometries = [_g.to(device) for _g in self.full_geometries]

        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)
        self.wavelengths_data = self.wavelengths_data.to(device)

        self.gpm_data = self.gpm_data.to(device)
        self.gpm_pos = self.gpm_pos.to(device)
        self.selfterms_data = self.selfterms_data.to(device)

        # transfer the lookup tables
        for wl in self.lookup:
            self.lookup[wl] = self.lookup[wl].to(device)

    def _interpolate_single_alpha(self, wls, gpm_data, wl_data):

        # convert to tensor
        wls = torch.as_tensor(wls, dtype=DTYPE_FLOAT, device=self.device)
        wl_dat = torch.as_tensor(wl_data, dtype=DTYPE_FLOAT, device=self.device)
        gpm_dat = torch.as_tensor(gpm_data, dtype=DTYPE_COMPLEX, device=self.device)

        gpm_size = gpm_dat.shape[1]
        assert gpm_dat.shape[1] == gpm_dat.shape[2]

        # wavelength-interpolation of each tensor component
        a_ip = torch.zeros(
            (len(wls), gpm_size, gpm_size), dtype=DTYPE_COMPLEX, device=self.device
        )

        for i in range(gpm_size):
            for j in range(gpm_size):
                _func_ip = interp.RegularGridInterpolator([wl_dat], gpm_dat[:, i, j])
                _ip = _func_ip([wls])
                a_ip[:, i, j] = _ip

        return a_ip

    def interpolate_alpha(self, wls, a_data_many, wl_data, lookup=None):
        """interpolate the polarizabilities between available wavelengths"""
        from torchgdm.tools.misc import tqdm

        gpm_ip = torch.zeros(
            (
                self.gpm_data.shape[0],
                len(wls),
                self.gpm_data.shape[2],
                self.gpm_data.shape[3],
            ),
            dtype=DTYPE_COMPLEX,
            device=self.device,
        )

        # iterate all polarizabilities (different structures)
        for i_a, a_data in tqdm(
            enumerate(a_data_many),
            title="creating GPM lookup...",
            progress_bar=self.progress_bar,
        ):
            _gpm = self._interpolate_single_alpha(wls, a_data, wl_data)
            gpm_ip[i_a] = _gpm

        # optionally add to lookup
        if lookup is not None:
            for i_wl, wl in enumerate(wls):
                wl = round(float(wl), 3)
                if wl not in lookup:
                    lookup[wl] = gpm_ip[:, i_wl, :, :]

        return gpm_ip

    # --- self-terms
    # self-terms are zero

    # --- polarizabilities
    def _call_interpolation(self, wavelength):
        warnings.warn(
            "Interpolating polarizabilities at wavelength {:.3f}.".format(
                float(wavelength)
            )
        )
        self.interpolate_alpha(
            [wavelength],
            self.gpm_data,
            self.wavelengths_data,
            lookup=self.lookup,
        )

    def get_gpm(self, wavelength: float, environment) -> torch.Tensor:
        """return list of GPM tensors (N_struct, 6N, 6N) of each GPM structure

        Args:
            wavelength (float): in nm

        Returns:
            list of torch.Tensor
        """
        if round(float(wavelength), 3) not in self.lookup:
            self._call_interpolation(wavelength)

        return self.lookup[round(float(wavelength), 3)]

    def get_polarizability_6x6(self, wavelength: float, environment) -> torch.Tensor:
        """return diagonal N6xN6 GPM

        warning: this does not return 6x6 tensors but
        GPMs of shape (N_structure, 6N, 6N)
        """
        return self.get_gpm(wavelength, environment)

    def get_selfterm_6x6(self, wavelength: float, environment) -> torch.Tensor:
        """return diagonal N6xN6 self-terms (zero)

        warning: this does not return 6x6 tensors but
        zeros of shape (N_structure, 6N, 6N)
        """
        return torch.zeros_like(self.gpm_data[:, 0])

    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        raise Exception("GPM does not provide single, local polarizabilities")

    def get_polarizability_mE(self, wavelength: float, environment) -> torch.Tensor:
        raise Exception("GPM does not provide single, local polarizabilities")

    def get_polarizability_pH(self, wavelength: float, environment) -> torch.Tensor:
        raise Exception("GPM does not provide single, local polarizabilities")

    def get_polarizability_mH(self, wavelength: float, environment) -> torch.Tensor:
        raise Exception("GPM does not provide single, local polarizabilities")

    # --- geometry operations
    def translate(self, vector):
        """return a copy moved by `vector`"""

        vector = torch.as_tensor(vector, dtype=DTYPE_FLOAT, device=self.device)
        vector = torch.atleast_2d(vector)
        _shifted = self.copy()

        # shift effective dipole positions of each GPM
        _shifted.gpm_pos = self.gpm_pos + vector.unsqueeze(0)

        if len(_shifted.full_geometries) > 0:
            for _g in _shifted.full_geometries:
                _g += vector

        # shift center of mass positions
        _shifted.r0 = _shifted.get_center_of_mass()

        return _shifted

    def set_center_of_mass(self, r0_new: torch.Tensor):
        """move center of mass to new position `r0_new` (in-place)"""
        r0_new = torch.as_tensor(r0_new, device=self.device)
        r0_old = self.get_center_of_mass()

        if len(r0_new.shape) != 1:
            if len(r0_new) not in [2, 3]:
                raise ValueError("`r0_new` needs to be (X,Y) or (X,Y,Z) tuple.")

        # treat case r0_new is 2D (X,Y)
        if len(r0_new) == 2:
            r0_new = torch.as_tensor(
                [r0_new[0], r0_new[1], r0_old[2]], device=self.device
            )

        # move each dipole to origin, then to new location
        self.gpm_pos -= (r0_old - r0_new).unsqueeze(0).unsqueeze(1)

        if len(self.full_geometries) > 0:
            # move to origin, the to new pos.
            self.full_geometries = [
                _g - (r0_old - r0_new) for _g in self.full_geometries
            ]

        self.r0 = self.get_center_of_mass()

    # --- plotting
    def plot(
        self,
        projection="auto",
        scale=1.0,
        color="auto",
        linestyle_circle=(0, (2, 2)),
        color_circle="auto",
        color_circle_fill=None,
        alpha=1,
        show_grid=True,
        color_grid="auto",
        alpha_grid=0.25,
        legend=True,
        set_ax_aspect=True,
        reset_color_cycle=True,
        gpm_plot_source_probes=False,
        **kwargs,
    ):
        """plot the point polarizability structure (2D)

        Args:
            projection (str, optional): Cartesian projection. Default: "XY" or plane in which all dipoles lie. Defaults to "auto".
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            linestyle_circle (tuple, optional): optional line style for enclosing circle. Defaults to (0, (2, 2)).
            color_circle (str, optional): optional alternative color for enclosing circle. Defaults to "auto".
            color_circle_fill (_type_, optional): optional alternative fill color for enclosing circle. Defaults to None.
            alpha (int, optional): optional transparency. Defaults to 1.
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            color_grid (str, optional): optional alternative color for the mesh grid. Defaults to "auto".
            alpha_grid (float, optional): optional alternative transparency for the mesh grid. Defaults to 0.25.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.
            gpm_plot_source_probes (bool, optional): Plot the GPM extraction source and probe locations. Requires `projection` specifically given. Defaults to False.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        if gpm_plot_source_probes and type(self.gpm_dict) == dict:
            import matplotlib.pyplot as plt
            from torchgdm.tools.misc import to_np

            # - probes
            r_prb = torch.stack([_r for _r in self.gpm_dict["extraction_r_probe"]])
            r_prb = to_np(r_prb + self.r0)
            if projection == "xy":
                ix, iy = 0, 1
            if projection == "xz":
                ix, iy = 0, 2
            if projection == "yz":
                ix, iy = 1, 2
            if projection in ["xy", "xz", "yz"]:
                plt.scatter(
                    r_prb[:, ix],
                    r_prb[:, iy],
                    color="g",
                    marker="o",
                    s=2,
                    label="GPM probes",
                )

            # - sources
            r_src = torch.stack(
                [
                    inc.r_source
                    for inc in self.gpm_dict["extraction_illuminations"]
                    if hasattr(inc, "r_source")
                ]
            )
            r_src = to_np(r_src + self.r0)
            if projection == "xy":
                ix, iy = 0, 1
            if projection == "xz":
                ix, iy = 0, 2
            if projection == "yz":
                ix, iy = 1, 2
            if projection in ["xy", "xz", "yz"]:
                plt.scatter(
                    r_src[:, ix],
                    r_src[:, iy],
                    color="r",
                    marker="x",
                    s=15,
                    label="GPM sources",
                )

        im = visu2d.geo2d._plot_structure_eff_pola(
            self,
            projection=projection,
            scale=scale,
            color=color,
            linestyle_circle=linestyle_circle,
            color_circle=color_circle,
            color_circle_fill=color_circle_fill,
            alpha=alpha,
            show_grid=show_grid,
            color_grid=color_grid,
            alpha_grid=alpha_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )

        return im

    def plot_contour(
        self,
        projection="auto",
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour of the underlying mesh (2D)

        Args:
            projection (str, optional): which cartesian plane to project onto. Defaults to "auto".
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu.visu2d._tools import _get_axis_existing_or_new
        from torchgdm.visu.visu2d.geo2d import _plot_contour_discretized
        from torchgdm.visu.visu2d.geo2d import _reset_color_iterator

        if reset_color_cycle:
            _reset_color_iterator()

        if len(self.full_geometries) == 0:
            warnings.warn("No mesh grid data available. Skipping.")
            return None
        ax, show = _get_axis_existing_or_new()
        for subgeo in self.full_geometries:
            im = _plot_contour_discretized(
                subgeo,
                projection=projection,
                color=color,
                alpha=alpha,
                alpha_value=alpha_value,
                set_ax_aspect=set_ax_aspect,
                **kwargs,
            )
        return im

    def plot3d(self, **kwargs):
        """plot the point polarizability structure (3D)"""
        from torchgdm.visu import visu3d

        return visu3d.geo3d._plot_structure_eff_3dpola(self, **kwargs)

    # --- geometry operations
    def rotate(self, alpha, center=torch.as_tensor([0.0, 0.0, 0.0]), axis="z"):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis, one of ['x', 'y', 'z']. Defaults to 'z'.

        Raises:
            ValueError: unknown rotation axis

        Returns:
            :class:`StructEffPola3D`: copy of structure with rotated geometry
        """
        _s_rot = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "x":
            rot = rotation_x(alpha, device=self.device)
        elif axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        elif axis.lower() == "z":
            rot = rotation_z(alpha, device=self.device)
        else:
            raise ValueError("Unknown rotation axis ''.".format(axis))

        # rotate GPM positions around `center`
        for i in range(len(_s_rot.gpm_pos)):
            _s_rot.gpm_pos[i] = torch.matmul(
                _s_rot.gpm_pos[i] - (center + self.r0), rot
            ) + (center + self.r0)

        # rotate full discretizations (if available)
        for i_g, _geo in enumerate(_s_rot.full_geometries):
            if len(_geo) > 1:
                _s_rot.full_geometries[i_g] = torch.matmul(
                    _geo - (center + self.r0), rot
                ) + (center + self.r0)
        # rotate GPM extraction probes and sources (if available)
        if type(_s_rot.gpm_dict) == dict:
            _s_rot.gpm_dict["extraction_r_probe"] = torch.matmul(
                _s_rot.gpm_dict["extraction_r_probe"], rot
            )
            for inc in _s_rot.gpm_dict["extraction_illuminations"]:
                if hasattr(inc, "r_source"):
                    inc.r_source = torch.matmul(inc.r_source, rot)

        # rotate all local and non-local polarizability tensors
        rot = rot.to(DTYPE_COMPLEX).unsqueeze(0).unsqueeze(0)
        rotT = rot.transpose(-2, -1)
        N_dp = 2 * self.gpm_pos.shape[1]  # rotate electric / magnetic dps separately
        for i_m in range(N_dp):
            m = 3 * i_m
            for i_n in range(N_dp):
                n = 3 * i_n
                _s_rot.gpm_data[..., m : m + 3, n : n + 3] = torch.matmul(
                    torch.matmul(
                        rotT,
                        _s_rot.gpm_data[..., m : m + 3, n : n + 3],
                    ),
                    rot,
                )

        # update lookup tables
        _s_rot.create_lookup()

        return _s_rot

    def combine(
        self, other, inplace=False, refresh_lookup=True, on_distance_violation="warn"
    ):
        """combine with a second structure

        Structures must be of same coupling type (electric / magnetic)

        Args:
            other (_type_): _description_
            inplace (bool, optional): Don't copy original structure, just add other structure. Can be necessary e.g. when gradients are required. Defaults to False.
            refresh_lookup (bool, optional): refresh the polarizability lookup table. Defaults to True.
            on_distance_violation (str, optional): can be "error", "warn", None (do nothing). Defaults to "error".

        Returns:
            :class:`StructBase`: new structure
        """
        if inplace:
            new_struct = self
        else:
            new_struct = self.copy()

        assert torch.all(new_struct.wavelengths_data == other.wavelengths_data)
        assert type(self) == type(other)

        N_dist1, N_dist2 = test_structure_distances(self, other)
        if on_distance_violation == "error" and (N_dist1 + N_dist2 > 0):
            raise ValueError(
                "Several meshpoints in structures are too close (struct1: {}, structu2: {})!".format(
                    N_dist1, N_dist2
                )
            )
        elif on_distance_violation == "warn" and (N_dist1 + N_dist2 > 0):
            warnings.warn(
                "Several meshpoints in structures are too close (struct1: {}, structu2: {})!".format(
                    N_dist1, N_dist2
                )
            )

        new_struct.gpm_pos = torch.concatenate([self.gpm_pos, other.gpm_pos], dim=0)
        new_struct.step = torch.concatenate([new_struct.step, other.step], dim=0)

        new_struct.gpm_data = torch.concatenate(
            [new_struct.gpm_data, other.gpm_data], dim=0
        )
        new_struct.selfterms_data = torch.concatenate(
            [new_struct.selfterms_data, other.selfterms_data], dim=0
        )

        # finally, add full geometries lists together
        new_struct.full_geometries = new_struct.full_geometries + other.full_geometries

        new_struct.r0 = new_struct.get_center_of_mass()
        # create lookup
        if refresh_lookup:
            new_struct.create_lookup()

        return new_struct


# --- Mie sphere - GPM3d
class StructMieSphereGPM3D(StructGPM3D):
    """class for Mie-theory based 3D GPM

    Defines a global polarizability matrix structure representing a core-shell sphere.
    Caution, GPM is an empirical model, its accuracy depends on a successful extraction procedure.

    Requires external package `treams`
    !!! class constructor does not support automatic differentiation !!!

    """

    __name__ = "Mie-theory sphere GPM (3D) structure class"

    def __init__(
        self,
        r_gpm: torch.Tensor,
        wavelengths: torch.Tensor,
        radii: list,
        materials: list,
        lmax=3,
        environment=None,
        r0: torch.Tensor = None,
        device: torch.device = None,
        verbose=True,
        progress_bar=True,
        **kwargs,
    ):
        """3D GPM class for a core-shell sphere (Mie)

        Use Mie theory to get an effective GPM model for a core-shell sphere.
        Requires the `treams` package for Mie coefficient calculation.
        https://github.com/tfp-photonics/treams

        `pip install treams`

        Args:
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            wavelengths (torch.Tensor): list of wavelengths to evaluate (nm)
            radii (list): list of the sphere's core and (multiple) shell radii (in nm).
            materials (list): materials of core and shell(s). A float or int is interpreted as permittivity value.
            lmax (int, optional): maximum order of Mie expansion. Defaults to 3.
            environment (environment instance, optional): Homogeneous 3D environment to evaluate Mie theory in. Defaults to None, which uses vacuum.
            r0 (torch.Tensor, optional): GPM structure position (x,y,z). If not given, is set to (0, 0, 0). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            verbose (bool, optional): whether to print progess info. Defaults to True.
            progress_bar (bool, optional): Progress bar for several tmatrices. Defaults to True.
            **kwargs: are passed to :func:`torchgdm.struct.eff_model_tools.extract_gpm_from_tmatrix`

        Raises:
            ValueError: incorrect parameters
        """
        import numpy as np
        from torchgdm.tools.misc import to_np
        from torchgdm.tools.misc import _check_environment
        from torchgdm.tools.misc import get_default_device
        from torchgdm.struct.eff_model_tools import extract_gpm_from_tmatrix
        from torchgdm.materials import MatConstant

        try:
            # ignore import warnings
            with warnings.catch_warnings():
                import treams
        except ModuleNotFoundError:
            print("Requires `treams`, install via `pip install treams`.")
            raise

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # environment
        env = _check_environment(environment, N_dim=3, device=device)

        # tensor conversion
        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)
        k0 = 2 * torch.pi / wavelengths
        k0 = torch.as_tensor(k0, device=device)

        # radii to array
        radii = np.atleast_1d(radii)
        r_enclosing = np.max(radii)  # outer radius

        # if single material, put in list
        if not hasattr(materials, "__iter__"):
            materials = [materials]

        n_env = np.zeros(len(wavelengths), dtype=np.complex128)
        tmatrix_list = []
        for i_wl, wl in enumerate(wavelengths):

            # embedding medium and wavevector therein
            eps_env = to_np(env.env_material.get_epsilon(wavelength=wl).real)[0, 0]
            n_env[i_wl] = eps_env**0.5

            # core and shell materials
            mat_treams = []
            for mat in materials:
                if type(mat) in [float, int, complex]:
                    mat = MatConstant(mat)
                eps_mat = to_np(mat.get_epsilon(wavelength=wl))[0, 0]
                mat_treams.append(treams.Material(eps_mat))

            # add environment material last
            mat_treams.append(treams.Material(eps_env))

            tmatrix_list.append(
                treams.TMatrix.sphere(lmax, to_np(k0[i_wl]), radii, mat_treams)
            )

        dict_gpm = extract_gpm_from_tmatrix(
            tmatrix_list,
            r_enclosing=r_enclosing,
            wavelengths=wavelengths,
            r_gpm=r_gpm,
            environment=env,
            device=self.device,
            verbose=verbose,
            progress_bar=progress_bar,
            **kwargs,
        )

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3

        super().__init__(positions=r0, gpm_dicts=[dict_gpm], device=self.device)


class StructTMatrixGPM3D(StructGPM3D):
    """class for T-Matrix based 3D GPM

    Defines a global polarizability matrix structure approximating T-Matrices.
    Caution, GPM is an empirical model, its accuracy depends on a successful extraction procedure.

    Requires external package `treams`
    !!! class constructor does not support automatic differentiation !!!

    """

    __name__ = "T-Matrix based GPM (3D) structure class"

    def __init__(
        self,
        tmatrices,
        r_gpm: torch.Tensor,
        r_enclosing: float,
        r0: torch.Tensor = None,
        device: torch.device = None,
        verbose=True,
        progress_bar=True,
        test_accuracy=True,
        **kwargs,
    ):
        """3D GPM class to approximate an arbitrary T-Matrix

        Use `treams` to extract an effective GPM model for a T-Matrix (or a list of).
        If a list is given, it needs to contain spectrally resolved T-Matrices (each at a different wavelength).
        Torchgdm assumes that the T-Matrices describe the same object.

        Requires the `treams` package for T-matrix scattered field calculations.
        https://github.com/tfp-photonics/treams

        `pip install treams`


        Args:
            tmatrices (list): list of treams T-matrices. Each T-Matrix must describe the same object at different wavelengths.
            r_enclosing (float): radius of T-matrix enclosing sphere in nm (assume center at origin).
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            r0 (torch.Tensor, optional): position to move the structure to (x,y,z). If not given, use the origin (0, 0, 0). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            verbose (bool, optional): whether to print progess info. Defaults to True.
            progress_bar (bool, optional): Progress bar for several tmatrices. Defaults to True.
            test_accuracy (bool, optional): Whether to test accuracy against a T-Matrix scattering simulation. Defaults to True.
            **kwargs: are passed to :func:`torchgdm.struct.eff_model_tools.extract_gpm_from_tmatrix`

        Raises:
            ValueError: incorrect parameters
        """
        from torchgdm.tools.misc import get_default_device
        from torchgdm.struct.eff_model_tools import extract_gpm_from_tmatrix
        from torchgdm.tools.tmatrix import _test_effective_model_accuracy_3d

        try:
            # ignore import warnings
            with warnings.catch_warnings():
                import treams
        except ModuleNotFoundError:
            print("Requires `treams`, install via `pip install treams`.")
            raise

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        dict_gpm = extract_gpm_from_tmatrix(
            tmatrices,
            r_enclosing=r_enclosing,
            r_gpm=r_gpm,
            device=self.device,
            verbose=verbose,
            progress_bar=progress_bar,
            **kwargs,
        )

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3

        super().__init__(positions=r0, gpm_dicts=[dict_gpm], device=self.device)

        if test_accuracy:
            _test_results = _test_effective_model_accuracy_3d(
                self, dict_gpm["t_matrices"]
            )



# -*- coding: utf-8 -*-
"""point polarizability classes"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.struct.struct3d.gpm3d import StructGPM3D
from torchgdm.struct.struct2d.surface import StructDiscretized2D
from torchgdm.tools.misc import ptp
from torchgdm.tools.misc import get_default_device


# --- base class volume discretized structure container - 3D
class StructGPM2D(StructGPM3D):
    """class for 2D line polarizability structure (infinite y axis)"""

    __name__ = "effective line polarizability (2D) structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        gpm_dicts: list,
        radiative_correction: bool = True,
        device: torch.device = None,
        environment=None,
        shift_z_to_r0: bool = True,
        progress_bar=True,
    ):
        """2D line polarizability class

        The main information is provided in the `alpha_dicts`, which is a list of dicts with the full effective polarizability definitions. Each dict defines one structure and must contain following:
            - 'wavelengths': wavelengths at which the polarizabilities are calculated
            - at least one of: ['alpha_pE', 'alpha_mH', 'alpha_mE', 'alpha_pH']:
                polarizability tensors of shape [len(wavelengths), 3, 3]
            optional keys:
            - 'full_geometry': the original volume discretization of the represented geometry
            - 'r0': the origin of the effective polarizabilities with respect to optional 'full_geometry'
            - 'enclosing_radius': enclosing radius of original structure

        Args:
            positions (torch.Tensor): polarizability positions (3D, but all y values must be zero)
            alpha_dicts (list): list of polarizability model dictionaries
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to True.
            device (torch.device, optional): Defaults to "cpu".
            environment (`environment` instance, optional): 2D environment class. By default use environment defined in eff.pola. dictionary. Defaults to None.
            shift_z_to_r0 (bool, optional): If True, if a position z-value is zero, each polarizability model's z position will be shifted to the height of the effective dipole development center. Defaults to True.

        Raises:
            ValueError: positions for 2D simulations must be at y=0
        """
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # expand positions, put single scatterer in list
        _positions = torch.as_tensor(positions, dtype=DTYPE_FLOAT, device=self.device)
        if len(_positions.shape) == 1:
            assert len(_positions) == 3
            _positions = _positions.unsqueeze(0)

        if torch.count_nonzero(_positions[..., 1]) > 0:
            warnings.warn("2D structure. Remove all positions with y!=0.")
            _positions = _positions[_positions[..., 1] != 0]
            if len(_positions) == 0:
                raise ValueError("No mesh positions at y=0. Please check geometry.")

        super().__init__(
            positions=positions,
            gpm_dicts=gpm_dicts,
            environment=environment,
            shift_z_to_r0=shift_z_to_r0,
            device=device,
            progress_bar=progress_bar,
        )

        self.n_dim = 2

        self.radiative_correction = radiative_correction

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ 2D effective line GPM object -------"
        out_str += "\n" + " nr. of dipole-pairs:    {}".format(
            len(self.get_all_positions())
        )
        out_str += "\n" + " nominal enclosing circle diameters (nm): {}".format(
            [round(float(f), 1) for f in torch.unique(self.step)]
        )
        if len(self.full_geometries) > 0:
            pos = torch.cat(self.full_geometries)
            out_str += "\n" + " original 2D geometry: "
            out_str += "\n" + "  - replacing nr. of meshpoints: {}".format(len(pos))
            bnds = ptp(pos, dim=0)
            out_str += "\n" + "  - size & position:"
            out_str += "\n" + "        X-extension    :    {:.1f} (nm)".format(bnds[0])
            out_str += "\n" + "        Z-extension    :    {:.1f} (nm)".format(bnds[2])
            out_str += "\n" + "  - center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
                *[float(f) for f in self.get_center_of_mass()]
            )

        return out_str

    # - radiative correction for cross section calc. - 2D case
    # inherit from discretized 2D class
    get_radiative_correction_prefactor_p = (
        StructDiscretized2D.get_radiative_correction_prefactor_p
    )
    get_radiative_correction_prefactor_m = (
        StructDiscretized2D.get_radiative_correction_prefactor_m
    )

    # --- plotting
    def plot(
        self,
        scale=1.0,
        color="auto",
        linestyle_circle=(0, (2, 2)),
        color_circle="auto",
        color_circle_fill=None,
        alpha=1,
        show_grid=True,
        color_grid="auto",
        alpha_grid=0.25,
        legend=True,
        set_ax_aspect=True,
        reset_color_cycle=True,
        gpm_plot_source_probes=False,
        **kwargs,
    ):
        """plot the structure of the effective line-polarizability (2D)

        Args:
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            linestyle_circle (tuple, optional): optional line style for enclosing circle. Defaults to (0, (2, 2)).
            color_circle (str, optional): optional alternative color for enclosing circle. Defaults to "auto".
            color_circle_fill (_type_, optional): optional alternative fill color for enclosing circle. Defaults to None.
            alpha (int, optional): optional transparency. Defaults to 1.
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            color_grid (str, optional): optional alternative color for the mesh grid. Defaults to "auto".
            alpha_grid (float, optional): optional alternative transparency for the mesh grid. Defaults to 0.25.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.
            gpm_plot_source_probes (bool, optional): Plot the GPM extraction source and probe locations. Defaults to False.

        Returns:
            matplotlib axes
        """
        kwargs["projection"] = "xz"
        im = super().plot(
            scale=scale,
            color=color,
            linestyle_circle=linestyle_circle,
            color_circle=color_circle,
            color_circle_fill=color_circle_fill,
            alpha=alpha,
            show_grid=show_grid,
            color_grid=color_grid,
            alpha_grid=alpha_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            reset_color_cycle=reset_color_cycle,
            gpm_plot_source_probes=gpm_plot_source_probes,
            **kwargs,
        )

        return im

    def plot_contour(
        self,
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour of the underlying 2D-mesh (2D)

        Args:
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        kwargs["projection"] = "xz"
        im = super().plot_contour(
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )

        return im

    def plot3d(self, **kwargs):
        """plot the point polarizability structure (3D)"""
        im = super().plot3d(**kwargs)

        return im

    # --- geometry operations
    def rotate(
        self,
        alpha: float,
        center: torch.Tensor = torch.as_tensor([0.0, 0.0, 0.0]),
        axis: str = "y",
    ):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis. Defaults to "y".

        Raises:
            ValueError: only "y" axis supported in 2D

        Returns:
            :class:`StructDiscretized2D`: copy of structure with rotated geometry
        """
        if axis.lower() != "y":
            raise ValueError(
                "Only rotation axis 'y' supported in 2D (infinite axis).".format(axis)
            )

        _struct_rotated = super().rotate(alpha=alpha, center=center, axis=axis)

        return _struct_rotated


# --- Mie sphere - GPM3d
class StructMieCylinderGPM2D(StructGPM2D):
    """class for Mie-theory based 2D GPM

    Defines a global polarizability matrix structure representing an infinite core-shell cylinder.
    TorchGDM infinite axis is along Y.
    Caution, GPM is an empirical model, its accuracy depends on a successful extraction procedure.

    Requires external package `treams`
    !!! class constructor does not support automatic differentiation !!!

    """

    __name__ = "Mie-theory cylinder GPM (2D) structure class"

    def __init__(
        self,
        r_gpm: torch.Tensor,
        wavelengths: torch.Tensor,
        radii: list,
        materials: list,
        mmax=3,
        environment=None,
        r0: torch.Tensor = None,
        device: torch.device = None,
        verbose=True,
        progress_bar=True,
        **kwargs,
    ):
        """2D GPM class for a core-shell cylinder (Mie)

        Use Mie theory to get an effective GPM model for a core-shell cylinder.
        Requires the `treams` package for Mie coefficient calculation.
        https://github.com/tfp-photonics/treams

        `pip install treams`

        Args:
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            wavelengths (torch.Tensor): list of wavelengths to evaluate (nm)
            radii (list): list of the sphere's core and (multiple) shell radii (in nm).
            materials (list): materials of core and shell(s). A float or int is interpreted as permittivity value.
            mmax (int, optional): maximum order of Mie expansion. Defaults to 3.
            environment (environment instance, optional): Homogeneous 2D environment to evaluate Mie theory in. Defaults to None, which uses vacuum.
            r0 (torch.Tensor, optional): GPM structure position (x,y,z). If not given, is set to (0, 0, 0). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            verbose (bool, optional): whether to print progess info. Defaults to True.
            progress_bar (bool, optional): Progress bar for several tmatrices. Defaults to True.
            **kwargs: are passed to :func:`torchgdm.struct.eff_model_tools.extract_gpm_from_tmatrix`

        Raises:
            ValueError: incorrect parameters
        """
        import numpy as np
        from torchgdm.tools.misc import to_np
        from torchgdm.tools.misc import _check_environment
        from torchgdm.tools.misc import get_default_device
        from torchgdm.struct.eff_model_tools import extract_gpm_from_tmatrix
        from torchgdm.materials import MatConstant

        try:
            # ignore import warnings
            with warnings.catch_warnings():
                import treams
        except ModuleNotFoundError:
            print("Requires `treams`, install via `pip install treams`.")
            raise

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # environment
        env = _check_environment(environment, N_dim=2, device=device)

        # tensor conversion
        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )
        wavelengths = torch.atleast_1d(wavelengths)
        k0 = 2 * torch.pi / wavelengths
        k0 = torch.as_tensor(k0, device=device)

        # radii to array
        radii = np.atleast_1d(radii)
        r_enclosing = np.max(radii)  # outer radius

        # if single material, put in list
        if not hasattr(materials, "__iter__"):
            materials = [materials]

        n_env = np.zeros(len(wavelengths), dtype=np.complex128)
        tmatrixC_list = []
        for i_wl, wl in enumerate(wavelengths):

            # embedding medium and wavevector therein
            eps_env = to_np(env.env_material.get_epsilon(wavelength=wl).real)[0, 0]
            n_env[i_wl] = eps_env**0.5

            # core and shell materials
            mat_treams = []
            for mat in materials:
                if type(mat) in [float, int, complex]:
                    mat = MatConstant(mat)
                eps_mat = to_np(mat.get_epsilon(wavelength=wl))[0, 0]
                mat_treams.append(treams.Material(eps_mat))

            # add environment material last
            mat_treams.append(treams.Material(float(eps_env)))

            tmatrixC_list.append(
                treams.TMatrixC.cylinder(
                    kzs=0.0,
                    mmax=mmax,
                    k0=float(to_np(k0[i_wl])),
                    radii=radii,
                    materials=mat_treams,
                )
            )

        dict_gpm = extract_gpm_from_tmatrix(
            tmatrixC_list,
            r_enclosing=r_enclosing,
            wavelengths=wavelengths,
            r_gpm=r_gpm,
            environment=env,
            device=self.device,
            verbose=verbose,
            progress_bar=progress_bar,
            **kwargs,
        )

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3

        super().__init__(positions=r0, gpm_dicts=[dict_gpm], device=self.device)


class StructTMatrixGPM2D(StructGPM2D):
    """class for cylindrical T-Matrix based 2D GPM

    Defines a global polarizability matrix structure approximating cylindrical (2D) T-Matrices.
    TorchGDM infinite axis is along Y.
    Caution, GPM is an empirical model, its accuracy depends on a successful extraction procedure.

    Requires external package `treams`
    !!! class constructor does not support automatic differentiation !!!

    """

    __name__ = "T-Matrix based GPM (2D) structure class"

    def __init__(
        self,
        tmatrices,
        r_gpm: torch.Tensor,
        r_enclosing: float,
        r0: torch.Tensor = None,
        device: torch.device = None,
        verbose=True,
        progress_bar=True,
        test_accuracy=True,
        **kwargs,
    ):
        """2D GPM class to approximate an arbitrary cylindrical T-Matrix (2D)

        Use `treams` to extract an effective GPM model for a 2D T-Matrix (or a list of).
        If a list is given, it needs to contain spectrally resolved T-Matrices (each at a different wavelength).
        Torchgdm assumes that the T-Matrices describe the same object.

        Requires the `treams` package for T-matrix scattered field calculations.
        https://github.com/tfp-photonics/treams

        `pip install treams`


        Args:
            tmatrices (list): list of treams T-matrices. Each T-Matrix must describe the same object at different wavelengths.
            r_enclosing (float): radius of T-matrix enclosing sphere in nm (assume center at origin).
            r_gpm (int or torch.Tensor): Either int: number of positions or a torch.Tensor of shape (N, 3), giving the locations where to place effective dipole pairs.
            r0 (torch.Tensor, optional): position to move the structure to (x,y,z). If not given, use the origin (0, 0, 0). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            verbose (bool, optional): whether to print progess info. Defaults to True.
            progress_bar (bool, optional): Progress bar for several tmatrices. Defaults to True.
            test_accuracy (bool, optional): Whether to test accuracy against a T-Matrix scattering simulation. Defaults to True.
            **kwargs: are passed to :func:`torchgdm.struct.eff_model_tools.extract_gpm_from_tmatrix`

        Raises:
            ValueError: incorrect parameters
        """
        from torchgdm.tools.misc import get_default_device
        from torchgdm.struct.eff_model_tools import extract_gpm_from_tmatrix
        from torchgdm.tools.tmatrix import _test_effective_model_accuracy_2d

        try:
            # ignore import warnings
            with warnings.catch_warnings():
                import treams
        except ModuleNotFoundError:
            print("Requires `treams`, install via `pip install treams`.")
            raise

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        dict_gpm = extract_gpm_from_tmatrix(
            tmatrices,
            r_enclosing=r_enclosing,
            r_gpm=r_gpm,
            device=self.device,
            verbose=verbose,
            progress_bar=progress_bar,
            **kwargs,
        )

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3

        super().__init__(positions=r0, gpm_dicts=[dict_gpm], device=self.device)

        if test_accuracy:
            _test_results = _test_effective_model_accuracy_2d(
                self, dict_gpm["t_matrices"]
            )
